(* Auxiliary definitions and lemmas *)


theory Aux imports Main

begin

(* Propositional logic *)

lemma mp_obj [simp]: "(A \<and> (A \<longrightarrow> B)) = (A \<and> B)"
by blast
lemma mp_obj2 [simp]: "((A \<longrightarrow> B) \<and> A) = (A \<and> B)"
by blast

-- "Absorption rules of the style A \<inter> (A \<union> B) = A"
lemmas absorb = Un_Int_distrib Un_Int_distrib2 Int_absorb1 Int_absorb2 Un_upper1 Un_upper2


(* Map *)

lemma dom_empty_map_empty [simp]: "(dom m = {}) = (m = empty)"
by simp

lemma map_le_implies_ran_le: "m \<subseteq>\<^sub>m m' \<Longrightarrow> ran m \<subseteq> ran m'"
by (fastforce simp add: dom_def ran_def map_le_def)

lemma image_dom: "m ` dom m = Some ` (ran m)"
by (auto simp add: image_def dom_def ran_def)

lemma image_dom_ran: "(the \<circ> m) ` dom m = ran m"
by (auto simp add: dom_def ran_def image_def)

lemma finite_dom_ran: "finite (dom m) \<Longrightarrow> finite (ran m)"
by (simp add: image_dom_ran [THEN sym])

lemma Some_the_map [simp]: "x \<in> dom m \<Longrightarrow> Some (the (m x)) = m x"
by (clarsimp simp add: dom_def)

lemma ran_map_upd_notin_dom: "a \<notin> dom m \<Longrightarrow> ran (m(a \<mapsto> b)) = ran m \<union> {b}"
by auto

lemma ran_map_upd_subset: "ran (m(a \<mapsto> b)) \<subseteq> ran m \<union> {b}"
by (auto simp add: ran_def)


lemma image_fun_upd: "
  (m ` ((f(x:=y)) e)) = (if e = x then m ` y else m ` (f e))"
by (clarsimp simp add: fun_upd_def image_def)


definition
  -- "injective map"
  inj_map :: "('a \<Rightarrow> 'b option) \<Rightarrow> bool" where
  "inj_map m == inj_on m (dom m)"


lemma inj_map_empty [simp]: "inj_map empty"
by (simp add: inj_map_def)

lemma subset_inj_map: "\<lbrakk> inj_map m; m' \<subseteq>\<^sub>m m \<rbrakk> \<Longrightarrow> inj_map m'"
apply (frule map_le_implies_dom_le)
apply (simp add: inj_map_def map_le_def inj_on_def)
apply blast
done

lemma inj_on_the: "inj_map gm \<Longrightarrow> inj_on (the \<circ> gm) (dom gm)"
apply (clarsimp simp add: inj_map_def inj_on_def)
apply (rename_tac x y v)
apply (drule_tac x=x in bspec) apply fastforce
apply (drule_tac x=y in bspec) apply fastforce
apply simp
done

lemma inj_on_map_upd:
  "a \<notin> dom m \<Longrightarrow> inj_on (m(a \<mapsto> b)) (dom m) =  inj_on m (dom m)"
apply (rule iffI)

apply (clarsimp simp add: inj_on_def)
apply (drule_tac bspec) prefer 2
apply (drule_tac bspec) prefer 2
apply (drule mp) prefer 2
apply assumption
apply (auto simp add: inj_on_def)
done

lemma inj_map_map_upd: "a \<notin> dom m \<Longrightarrow> inj_map (m(a \<mapsto> b)) = (b \<notin> ran m \<and> inj_map m)"
by (simp add: inj_map_def inj_on_map_upd image_dom) auto


(* Inversion of a map *)
definition
  inv_m :: "('a \<Rightarrow> 'b option) \<Rightarrow> ('b \<Rightarrow> 'a option)" where
  "inv_m m == \<lambda> y. if (y \<in> ran m) then Some (SOME x. m x = Some y) else None"


lemma inv_m_empty [simp]: "inv_m empty = empty"
by (simp add: inv_m_def)

lemma dom_inv_m [simp]: "dom (inv_m m) = ran m"
by (simp add: inv_m_def ran_def dom_def)

lemma ran_inv_m [simp]: "ran (inv_m m) \<subseteq>  dom m"
by (simp add: inv_m_def ran_def dom_def) (fast intro: someI2)

lemma restrict_map_le [simp]: "m |` A \<subseteq>\<^sub>m m"
by (simp add: map_le_def)

lemma restrict_map_le_in_dom: "\<lbrakk> m \<subseteq>\<^sub>m m'; dom m \<subseteq> A \<rbrakk> \<Longrightarrow> m \<subseteq>\<^sub>m m' |` A"
by (fastforce simp add: map_le_def )

lemma restrict_map_Some: "((m |` A) x = Some y) = (m x = Some y \<and> x \<in> A)"
by (simp add: restrict_map_def)


lemma inv_m_map_upd: "\<lbrakk> a \<notin> dom m; inj_map (m(a \<mapsto> b))\<rbrakk> \<Longrightarrow> 
  inv_m (m(a \<mapsto> b)) = (inv_m m)(b \<mapsto> a)"
apply (simp add: inv_m_def ran_map_upd_notin_dom)
apply (simp add: fun_eq_iff)
apply (intro allI impI conjI)

apply (simp add: inj_map_map_upd)
defer
apply (simp add: inj_map_map_upd)
apply (rule some_equality) apply simp apply (simp add: ran_def)

apply (rule_tac f=Eps in arg_cong)
apply (simp add: fun_eq_iff)
apply auto
done

lemma o_m_inv_m_reduce: 
  "\<lbrakk> a \<notin> dom m2; inj_map (m2(a \<mapsto> b))\<rbrakk> \<Longrightarrow> 
  m1 \<circ>\<^sub>m (inv_m (m2(a \<mapsto> b))) = (m1 \<circ>\<^sub>m inv_m m2)(b:= m1 a)"
apply (simp add: inv_m_map_upd)
apply (simp add: map_comp_def)
apply (simp add: fun_eq_iff)
done


(* restriction of a map *)
lemma restrict_map_le_eq: "((m|`A)  \<subseteq>\<^sub>m (m|`B)) = (dom m \<inter> A \<subseteq> B)"
apply (rule iffI)
apply (frule map_le_implies_dom_le) apply (fastforce simp only: dom_restrict) 
apply (simp add: map_le_def) apply (fastforce simp add: restrict_map_def)
done

lemma restrict_map_dom [simp]: "m \<subseteq>\<^sub>m m' \<Longrightarrow> m' |` (dom m) = m"
apply (rule map_le_antisym)
apply (simp add: restrict_map_def map_le_def dom_def)
apply (erule restrict_map_le_in_dom) apply simp
done

lemma restrict_map_dom_subset: "dom m \<subseteq> A \<Longrightarrow> m |` A = m"
apply (rule  map_le_antisym)
apply simp
apply (simp add: restrict_map_le_in_dom)
done

lemma ran_restrict_iff: "(y \<in> ran (m |` A)) = (\<exists>x\<in>A. m x = Some y)"
apply (rule iffI)
apply (erule ran_restrictD)
apply (auto simp add: ran_def restrict_map_def)
done


(* Relation *)

lemma Field_insert [simp]: "Field (insert (a,a') A) = {a,a'} \<union> Field A"
by (fastforce simp add: Field_def Domain_def Range_def)

lemma finite_Image: "finite (Field R) \<Longrightarrow> finite (R `` S)"
apply (rule finite_subset [where B="Range R"])
apply (auto simp add: Field_def)
done

lemma Image_Field_subset [simp]: "(R `` S) \<subseteq> Field R"
by (auto simp add: Field_def)

lemma Field_converse [simp]: "Field (R^-1) = Field R"
by (auto simp add: Field_def)

lemma Field_Un: "Field (R \<union> S) = (Field R  \<union> Field S)"
by (auto simp add: Field_def)

lemma Field_prod [simp]: "Field (A \<times> A) = A"
by (fastforce simp add: Field_def)

lemma Field_product_subset: "(A \<subseteq> B \<times> B) = (Field A \<subseteq> B)"
by (fastforce simp add: Field_def)

lemma Field_Diff_subseteq: "Field (R - S) \<subseteq> Field R"
by (auto simp add: Field_def)

lemma in_Field: "(a, b) \<in> R \<Longrightarrow> {a,b} \<subseteq> Field R"
by (fastforce simp add: Field_def Domain_def Range_def)

lemma converse_empty_set [simp]: "{}\<inverse> = {}"
by (simp add: converse_def)

lemma converse_insert: "(insert (x, y) R)\<inverse> = insert (y, x) (R\<inverse>)"
by (unfold insert_def) (auto simp add: converse_Un)

lemma converse_Diff: "(R - S)\<inverse> = R\<inverse> -  S\<inverse>"
by auto

lemma Diff_Image: "(R - S) `` {x} = R `` {x} - S `` {x}"
by blast

lemma finite_rel_finite_Field: "\<lbrakk> Field R \<subseteq> A; finite A \<rbrakk> \<Longrightarrow> finite R"
apply (subgoal_tac "R \<subseteq> A \<times> A")
apply (rule finite_subset) apply assumption+
apply (fast intro: finite_cartesian_product)
apply (fastforce simp add: Field_def)
done

lemma insert_Image_split:
  "((insert (a, b) R) `` {c}) = ((if a = c then {b}  else {}) \<union> (R `` {c}))"
by auto


lemma Image_rel_empty [simp]: "{} `` A = {}"
by auto

lemma converse_mono: "A \<subseteq> B \<Longrightarrow> (A)\<inverse> \<subseteq> (B)\<inverse>"
by auto

definition restrict_rel :: "[('a * 'a) set, 'a set] \<Rightarrow> ('a * 'a) set" where
  "restrict_rel R A = R \<inter> A \<times> A"

lemma Field_restrict_rel [simp]: "Field (restrict_rel R A) \<subseteq> A"
by (auto simp add: restrict_rel_def Field_def)

lemma restrict_rel_empty [simp]: "restrict_rel {} A = {}"
by (simp add: restrict_rel_def) 

lemma restrict_rel_insert:
  "(restrict_rel (insert (x,y) R) A) = 
    (if (x \<in> A \<and> y \<in> A) then {(x, y)} else {}) \<union> (restrict_rel R A)"
by (simp add: restrict_rel_def) 

lemma restrict_rel_insert2:
  "(restrict_rel (insert (x,y) R) A) = (({(x, y)} \<inter> A \<times> A) \<union> (restrict_rel R A))"
by (simp add: restrict_rel_def) blast

lemma restrict_rel_insert_dom:
  "restrict_rel R A = R \<Longrightarrow> restrict_rel R (insert a A) = R"
by (fastforce simp add: restrict_rel_def)

lemma restrict_relD [dest]: 
  "(a, a') \<in> restrict_rel R A \<Longrightarrow> (a,a') \<in> R \<and> a \<in> A \<and> a' \<in> A"
by (simp add: restrict_rel_def)

lemma restrict_rel_Diff: "restrict_rel (r - s) A = (restrict_rel r A) - (restrict_rel s A)"
by (fastforce simp add: restrict_rel_def)

lemma restrict_rel_Un [simp]:  
  "restrict_rel (R \<union> S) A = restrict_rel R A \<union> restrict_rel S A"
apply (simp add: restrict_rel_def)
apply blast
done

lemma restrict_rel_mono: "R \<subseteq> R' \<Longrightarrow> A \<subseteq> A' \<Longrightarrow> restrict_rel R A \<subseteq> restrict_rel R' A'"
by (fastforce simp add: restrict_rel_def)

lemma restrict_rel_Field_subset [simp]: 
  "Field R \<subseteq> A \<Longrightarrow> restrict_rel R A = R"
by (simp add: restrict_rel_def Field_def) fast


definition
  fun_map_upd :: "('a => 'b)  => ('a ~=> 'b) => ('a => 'b)" where
  "fun_map_upd f m  = (\<lambda>k. case m k of None \<Rightarrow> f k | Some v \<Rightarrow> v)"

lemma fun_map_upd_empty [simp]: "fun_map_upd f empty = f"
by (simp add: fun_map_upd_def)

lemma fun_map_upd_upd [simp]: "fun_map_upd f (m(x\<mapsto>y)) = (fun_map_upd f m)(x:=y)"
by (simp add: fun_map_upd_def fun_eq_iff)



lemma map_add_dom_disj1: "\<lbrakk>dom tt1 \<inter> dom tt2 = {}; tt1 b = Some ntp' \<rbrakk> 
  \<Longrightarrow> (tt1 ++ tt2) b = Some ntp'"
by (auto simp add: map_add_Some_iff)

lemma map_add_dom_disj2: "\<lbrakk>dom tt1 \<inter> dom tt2 = {}; tt2 b = Some ntp' \<rbrakk> 
  \<Longrightarrow> (tt1 ++ tt2) b = Some ntp'"
by (auto simp add: map_add_Some_iff)

lemma map_add_disj:
  "\<lbrakk> (m1 ++ m2) x = Some x'; dom m1 \<inter> dom m2 = {} \<rbrakk> \<Longrightarrow> (m1 x = Some x') \<or> (m2 x = Some x')"
by auto


lemma dom_map_comp: "ran g \<subseteq> dom f \<Longrightarrow> dom (f o_m g) = dom g"
apply (simp add: dom_def ran_def map_comp_def)
apply (auto split add: option.split_asm)
done


lemma map_add_image_ran: "dom m2 = A \<Longrightarrow>(the \<circ> m1 ++ m2) ` A = ran m2"
by (clarsimp simp add: map_add_def image_def ran_def dom_def) auto

lemma map_add_image: "dom m2 \<inter> A = {} \<Longrightarrow>(m1 ++ m2) ` A = (m1) ` A"
by (fastforce simp add: map_add_def image_def split: option.splits)+

lemma the_map_add_image: "dom m2 \<inter> A = {} \<Longrightarrow>(the \<circ> (m1 ++ m2)) ` A = (the \<circ> m1) ` A"
by (simp add: image_compose map_add_image)

lemma image_Int_subset: "A \<subseteq> B \<Longrightarrow> f ` (A \<inter> B) = f ` A \<inter> f ` B"
by (fastforce simp add: image_def)


lemma dom_reduce_insert: "
  (dom gm = insert a A) = (\<exists> b gm'. gm = gm'(a\<mapsto>b) \<and> dom gm' = A)"
apply (rule iffI)
apply (rule_tac x="the (gm a)" in exI)
apply (rule_tac x="gm|` A" in exI)

apply (simp add: restrict_map_insert [THEN sym] restrict_map_dom_subset)
apply fastforce
apply clarsimp
done


  (* Function map_pair *)


lemma inj_map_pair [simp]: "inj f \<Longrightarrow> inj g \<Longrightarrow> inj (map_pair f g)"
by (simp add: map_pair_def inj_on_def)

lemma rtrancl_map_pair: "(b0, c0) \<in> B\<^sup>* \<Longrightarrow> (f b0, f c0)  \<in> (map_pair f f ` B)\<^sup>*"
apply (induct b0 c0 rule: rtrancl.induct)
apply fast
apply (subgoal_tac "(f b, f c) \<in> (map_pair f f ` B)")
apply (rule rtrancl_into_rtrancl) apply assumption+
apply (fastforce simp add: map_pair_def image_def)
done

lemma rtrancl_inclusion_map_pair: 
  "A^* \<subseteq> B^* \<Longrightarrow>  ((map_pair f f) ` A)^* \<subseteq> ((map_pair f f) ` B)^*"
apply clarify
apply (rename_tac x y)
apply (erule_tac rtrancl.induct [where P = "\<lambda> x y.  (x, y) \<in>  ((map_pair f f) ` B)^*"])
apply fast
apply (subgoal_tac "\<exists> b0 c0. (b0, c0) \<in> A \<and> b = (f b0) \<and> c = (f c0)")
  prefer 2 apply (fastforce simp add: map_pair_def image_def)
apply clarsimp
apply (subgoal_tac "(b0, c0) \<in> B^*") prefer 2 apply blast
apply (fast intro: rtrancl_trans rtrancl_map_pair)
done

definition
  emorph :: "('a \<Rightarrow> 'b option) \<Rightarrow> 'a \<times> 'a \<Rightarrow> 'b \<times> 'b" where
  "emorph m = map_pair (the \<circ> m) (the \<circ> m)"

lemma emorph_pair [simp]: "emorph m (a, b) = (the (m a), the (m  b))"
by (simp add: emorph_def)

lemma emorph_converse:
 "emorph gm ` (R)\<inverse> = (emorph gm ` R)\<inverse>"
apply (simp add: emorph_def converse_def image_def split_beta)
apply (rule Collect_cong)
apply (rule iffI)
apply (clarsimp) apply (rule bexI) prefer 2 apply assumption apply simp
apply (clarsimp) apply fast
done

lemma Field_emorph: "Field (emorph m ` R) = (the \<circ>  m) ` (Field R)"
by (fastforce simp add: emorph_def image_def Field_def Domain_def Range_def)

lemma rtrancl_inclusion_emorph: 
  "A^* \<subseteq> B^* \<Longrightarrow>  ((emorph gm) ` A)^* \<subseteq> ((emorph gm) ` B)^*"
by (simp add: emorph_def rtrancl_inclusion_map_pair)

lemma inj_on_emorph:
  "inj_map gm \<Longrightarrow> inj_on (emorph gm) (dom gm \<times> dom gm)"
by (simp add: emorph_def map_pair_inj_on inj_on_the) 


  (* Transitive closure *)

lemma rtrancl_id_trancl: "R\<^sup>* = Id \<union> R\<^sup>+"
apply (simp add: set_eq_iff rtrancl_eq_or_trancl)
apply blast
done

end
