theory Decomposition imports Aux

begin


definition "interior_rel A r = r \<inter> (A \<times> A)"

definition "exterior_rel A r = r - (A \<times> A)"

lemma interior_rel_mono:  "A \<subseteq> A' \<Longrightarrow> r \<subseteq> r' \<Longrightarrow> interior_rel A r \<subseteq> interior_rel A' r'"
by (fastforce simp add: interior_rel_def)

  (* note: anti-mono in first argument *)
lemma exterior_rel_mono:  "A' \<subseteq> A \<Longrightarrow> r \<subseteq> r' \<Longrightarrow> exterior_rel A r \<subseteq> exterior_rel A' r'"
by (fastforce simp add: exterior_rel_def)


lemma interior_union_exterior [simp]:
  "interior_rel A r \<union> exterior_rel A r = r"
by (auto simp add: interior_rel_def exterior_rel_def)

lemma interior_disjoint_exterior [simp]:
  "interior_rel A r \<inter> exterior_rel A r = {}"
by (auto simp add: interior_rel_def exterior_rel_def)

lemma interior_exterior_decompos_rtrancl [simp]:
  "((interior_rel A r)\<^sup>* \<union> (exterior_rel A r)\<^sup>*)\<^sup>* = r\<^sup>*"
by (simp add: rtrancl_Un_rtrancl)

  (* Unfortunate naming of a rule trancl_mono in the Isabelle library *)
lemma trancl_mono_rel: "r \<subseteq> s \<Longrightarrow> r\<^sup>+ \<subseteq> s\<^sup>+"
by (fast intro: trancl_mono)

lemma trancl_Un_trancl [simp]: "(R\<^sup>+ \<union> S\<^sup>+)\<^sup>+ = (R \<union> S)\<^sup>+"
apply (rule subset_antisym)
apply clarsimp
apply (erule trancl.induct)
apply (fastforce intro: trancl_mono)
apply (fastforce intro: trancl_mono elim: trancl_trans)
apply (fastforce intro: trancl_mono)
done

lemma interior_exterior_decompos_trancl [simp]:
  "((interior_rel A r)\<^sup>+ \<union> (exterior_rel A r)\<^sup>+)\<^sup>+ = r\<^sup>+"
by simp


lemma Field_interior_rel: "Field r \<subseteq> A \<Longrightarrow> (interior_rel A r) = r"
by (fastforce simp add: interior_rel_def Field_def)

lemma Field_exterior_rel: "Field r \<subseteq> A \<Longrightarrow> (exterior_rel A r) = {}"
by (fastforce simp add: exterior_rel_def Field_def)


lemma interior_exterior_decompos_subset_equal:
  "((exterior_rel A r) \<subseteq> (exterior_rel A s) \<and>
    (interior_rel A r) \<subseteq> (interior_rel A s)) 
  = (r \<subseteq>  s)"
apply (rule iffI)
apply (subgoal_tac 
  "((interior_rel A r) \<union> (exterior_rel A r)) \<subseteq>
    ((interior_rel A s) \<union> (exterior_rel A s))")
apply simp
apply blast

apply (simp add: exterior_rel_mono interior_rel_mono)
done

lemma interior_exterior_decompos_subset:
  "((exterior_rel A r) \<subseteq> (exterior_rel A s) \<and>
    (interior_rel A r) \<subseteq> (interior_rel A s)) \<Longrightarrow> (r \<subseteq>  s)"
by (simp only: interior_exterior_decompos_subset_equal)


lemma interior_exterior_decompos_rtrancl_subset:
  "\<lbrakk> (exterior_rel A r)\<^sup>* \<subseteq> (exterior_rel A s)\<^sup>*; 
  (interior_rel A r)\<^sup>* \<subseteq> (interior_rel A s)\<^sup>* \<rbrakk> \<Longrightarrow>
  r\<^sup>* \<subseteq>  s\<^sup>*"
apply (subgoal_tac 
  "((interior_rel A r)\<^sup>* \<union> (exterior_rel A r)\<^sup>*)\<^sup>*
  \<subseteq>
  ((interior_rel A s)\<^sup>* \<union> (exterior_rel A s)\<^sup>*)\<^sup>*")
apply simp
apply (rule rtrancl_mono)
apply blast
done

lemma interior_exterior_decompos_trancl_subset:
  "\<lbrakk> (exterior_rel A r)\<^sup>+ \<subseteq> (exterior_rel A s)\<^sup>+; 
  (interior_rel A r)\<^sup>+ \<subseteq> (interior_rel A s)\<^sup>+ \<rbrakk> \<Longrightarrow>
  r\<^sup>+ \<subseteq>  s\<^sup>+"
apply (subgoal_tac 
  "((interior_rel A r)^+ \<union> (exterior_rel A r)^+)^+
  \<subseteq>
  ((interior_rel A s)^+ \<union> (exterior_rel A s)^+)^+")
apply simp
apply (rule trancl_mono_rel)
apply blast
done

  (* the converse of the implication below does not hold in general ... *)


  (* ... but in special cases:  *)

lemma decompos_complete_interior:
  "Field s \<subseteq> A \<Longrightarrow> r\<^sup>* \<subseteq>  s\<^sup>* \<Longrightarrow> (interior_rel A r)\<^sup>* \<subseteq> (interior_rel A s)\<^sup>*"
apply (simp add: Field_interior_rel)
apply (subgoal_tac "(interior_rel A r)\<^sup>* \<subseteq>  r\<^sup>*") apply fast
apply (rule rtrancl_mono)
apply (fastforce simp add: interior_rel_def)
done

lemma decompos_complete_exterior:
  "Field r \<subseteq> A \<Longrightarrow> r\<^sup>* \<subseteq>  s\<^sup>* \<Longrightarrow> (exterior_rel A r)\<^sup>* \<subseteq> (exterior_rel A s)\<^sup>*"
by (fastforce simp add: Field_exterior_rel)


lemma interior_exterior_decompos_rtrancl_elems:
  "\<lbrakk> (x, y) \<in> r\<^sup>*; 
  (exterior_rel A r)\<^sup>* \<subseteq> (exterior_rel A s)\<^sup>*; 
  (interior_rel A r)\<^sup>* \<subseteq> (interior_rel A s)\<^sup>* \<rbrakk> \<Longrightarrow> (x, y) \<in> s\<^sup>*"
by (blast dest: interior_exterior_decompos_rtrancl_subset)



text {* Simplification rules for exterior_rel with set-theoretic operators *}

lemma exterior_rel_empty [simp]: 
  "exterior_rel A {} = {}"
by (fastforce simp add: exterior_rel_def)

lemma exterior_rel_insert [simp]: 
  "exterior_rel A (insert (x, y) r) = ({(x, y)} - (A \<times> A)) \<union>  exterior_rel A r"
by (fastforce simp add: exterior_rel_def)

lemma exterior_rel_inter [simp]: 
  "exterior_rel A (r \<inter> s) = (exterior_rel A r) \<inter> (exterior_rel A s)"
by (fastforce simp add: exterior_rel_def)

lemma exterior_rel_union [simp]: 
  "exterior_rel A (r \<union> s) = (exterior_rel A r) \<union> (exterior_rel A s)"
by (fastforce simp add: exterior_rel_def)

lemma exterior_rel_diff [simp]: 
  "exterior_rel A (r - s) = (exterior_rel A r) - (exterior_rel A s)"
by (fastforce simp add: exterior_rel_def)

lemma exterior_rel_converse [simp]: 
  "exterior_rel A (r^-1) = (exterior_rel A r)^-1"
by (fastforce simp add: exterior_rel_def)

  (* Does not hold:
  "exterior_rel A (- s) =  - (exterior_rel A s)"
  *)

  (* not added to simpset because usefulness is questionable *)
lemma exterior_rel_rel_comp: 
  "exterior_rel A (r O s) = ((((exterior_rel A r) O (exterior_rel A s)) - (interior_rel A (r O s)))
                           \<union> ((exterior_rel A r) O (interior_rel A s))
                           \<union> ((interior_rel A r) O (exterior_rel A s)))"
by (fastforce simp add: rel_comp_def exterior_rel_def interior_rel_def)



text {* Simplification rules for interior_rel with set-theoretic operators *}

lemma "x \<in> A \<and> y \<in> A \<Longrightarrow> ({(x, y)} \<inter> (A \<times> A)) = {(x, y)}"
by simp

lemma "x \<notin> A \<or> y \<notin> A \<Longrightarrow> ({(x, y)} \<inter> (A \<times> A)) = {}"
by fastforce

lemma interior_rel_empty [simp]:
  "interior_rel A {} = {}"
by (fastforce simp add: interior_rel_def)

lemma interior_rel_insert [simp]: 
  "interior_rel A (insert (x, y) r) =  ({(x, y)} \<inter> (A \<times> A)) \<union> (interior_rel A r)"
by (fastforce simp add: interior_rel_def)

lemma interior_rel_inter [simp]: 
  "interior_rel A (r \<inter> s) = (interior_rel A r) \<inter> (interior_rel A s)"
by (fastforce simp add: interior_rel_def)

lemma interior_rel_union [simp]: 
  "interior_rel A (r \<union> s) = (interior_rel A r) \<union> (interior_rel A s)"
by (fastforce simp add: interior_rel_def)

lemma interior_rel_diff [simp]: 
  "interior_rel A (r - s) = (interior_rel A r) - (interior_rel A s)"
by (fastforce simp add: interior_rel_def)

lemma interior_rel_converse [simp]: 
  "interior_rel A (r^-1) = (interior_rel A r)^-1"
by (fastforce simp add: interior_rel_def)


lemma interior_rel_comp1 [simp]: 
  "((interior_rel A r) O (interior_rel A s)) \<subseteq> interior_rel A (r O s)"
by (fastforce simp add: interior_rel_def)



lemma convert_union: "(x, y) \<in> (r \<union> s) = ((x,y) \<in> r  \<or> (x,y) \<in> s)"
by blast

definition
  rel_comp_restricted  :: "['b set, ('a * 'b) set, ('b * 'c) set] => ('a * 'c) set" where
  "rel_comp_restricted B r s = {(x,z). \<exists> y \<in> B. (x, y) \<in> r \<and> (y, z) \<in>s}"

lemma convert_rel_comp: "((x, z) \<in> rel_comp_restricted B r s) = (\<exists> y \<in> B. (x, y) \<in> r \<and> (y, z) \<in> s)"
by (fastforce simp add: rel_comp_restricted_def)


text {* Simplification rules for interior_rel recursing over node set *}

definition "interior_rel_elem_r a B r = r \<inter> ({a} \<times> B)"
definition "interior_rel_elem_l b A r = r \<inter> (A \<times> {b})"

lemma interior_rel_empty_dom [simp]:
  "interior_rel {} r = {}"
by (simp add: interior_rel_def)

lemma interior_rel_set_insert:
  "interior_rel (insert a A) r = 
  ((interior_rel A r) \<union> (interior_rel_elem_r a (insert a A) r) \<union> (interior_rel_elem_l a (insert a A) r))"
by (fastforce simp add: interior_rel_def interior_rel_elem_r_def interior_rel_elem_l_def)

lemma interior_rel_elem_r_insert:
  "interior_rel_elem_r a (insert b B) r = (r \<inter> ({a} \<times> {b})) \<union> interior_rel_elem_r a B r"
by (fastforce simp add: interior_rel_elem_r_def)

lemma interior_rel_elem_r_insert_if:
  "interior_rel_elem_r a (insert b B) r = 
   ((if (a, b) \<in> r then {(a, b)} else {}) \<union> interior_rel_elem_r a B r)"
by (simp add: interior_rel_elem_r_def)

  (* Alternative version to allow for nesting of if expressions. 
     Not decisively more efficient than the above. *)
lemma interior_rel_elem_r_insert_if2:
  "interior_rel_elem_r a (insert b B) r = 
   (if (a, b) \<in> r then (insert (a, b) (interior_rel_elem_r a B r)) else (interior_rel_elem_r a B r))"
by (simp add: interior_rel_elem_r_def)

lemma interior_rel_elem_r_empty: "interior_rel_elem_r a {} r = {}"
by (fastforce simp add: interior_rel_elem_r_def)

lemma interior_rel_elem_l_insert:
  "interior_rel_elem_l b (insert a A) r = (r \<inter> ({a} \<times> {b})) \<union> interior_rel_elem_l b A r"
by (fastforce simp add: interior_rel_elem_l_def)

lemma interior_rel_elem_l_insert_if:
  "interior_rel_elem_l b (insert a A) r = 
   ((if (a, b) \<in> r then {(a, b)} else {}) \<union> interior_rel_elem_l b A r)"
by (simp add: interior_rel_elem_l_def)


lemma interior_rel_elem_l_insert_if2:
  "interior_rel_elem_l b (insert a A) r = 
   (if (a, b) \<in> r then (insert (a, b) (interior_rel_elem_l b A r)) else (interior_rel_elem_l b A r))"
by (simp add: interior_rel_elem_l_def)

lemma interior_rel_elem_l_empty: "interior_rel_elem_l b {} r = {}"
by (fastforce simp add: interior_rel_elem_l_def)


lemma "(insert (y, x) r)\<^sup>* = r\<^sup>* \<union> {(a, b). (a, y) \<in> r\<^sup>* \<and> (x, b) \<in> r\<^sup>*}"
by (fastforce simp add: set_eq_iff rtrancl_eq_or_trancl trancl_insert)

lemma pair_insert_rtrancl: "(v, w) \<in> (insert (y, x) r)\<^sup>* = ((v, w) \<in> r\<^sup>* \<or> (v, y) \<in> r\<^sup>* \<and> (x, w) \<in> r\<^sup>*)"
by (fastforce simp add: set_eq_iff rtrancl_eq_or_trancl trancl_insert)

lemma pair_insert_trancl: "(v, w) \<in> (insert (y, x) r)\<^sup>+ = 
  ((v, w) \<in> r\<^sup>+ \<or> ((v = y) \<or> (v, y) \<in> r\<^sup>+) \<and> ((x = w) \<or> (x, w) \<in> r\<^sup>+))"
by (fastforce simp add: set_eq_iff trancl_insert rtrancl_id_trancl)

  (* gives more compact terms during simplification: *)
lemma pair_insert_trancl2: "(v, w) \<in> (insert (y, x) r)\<^sup>+ = 
  ((v, w) \<in> r\<^sup>+ \<or> ((v \<noteq> y) \<longrightarrow> (v, y) \<in> r\<^sup>+) \<and> ((x \<noteq> w) \<longrightarrow> (x, w) \<in> r\<^sup>+))"
by (fastforce simp add: set_eq_iff trancl_insert rtrancl_id_trancl)

lemma trancl_insert_equational: 
  "(insert (y, x) r)\<^sup>+ = r\<^sup>+ \<union> (r\<^sup>* O {(y,x)} O r\<^sup>*)"
apply (rule subset_antisym)
apply (clarsimp simp add: subset_eq pair_insert_trancl rel_comp_def rtrancl_eq_or_trancl)
apply (elim disjE)
apply clarsimp+
apply (clarsimp simp add: subset_eq pair_insert_trancl rel_comp_def rtrancl_eq_or_trancl)
done


lemma rel_comp_rtrancl_right: "r' O r\<^sup>*  = r' \<union> r' O r\<^sup>+"
by (simp add: rtrancl_id_trancl) 
lemma rel_comp_rtrancl_left: "r\<^sup>* O r'  = r' \<union> r\<^sup>+ O r'"
by (simp add: rtrancl_id_trancl) 

lemma rel_comp_insert: "T \<noteq> {} \<Longrightarrow> R O (insert s T) = R O {s} \<union> R O T"
by (fastforce simp add: rel_comp_def)
lemma rel_comp_insert2: "T \<noteq> {} \<Longrightarrow> (insert s T) O R = {s} O R \<union> T O R"
by (fastforce simp add: rel_comp_def)

lemma rel_comp_insert_singleton: "{(x,v)} O {(w, y)} = (if v = w then {(x,y)} else {})"
by (simp add: rel_comp_def)


lemmas rel_comp_simplif = rel_comp_insert rel_comp_insert2 rel_comp_insert_singleton 


lemma "(v, w) \<in> ({})\<^sup>* = (v = w)"
by simp

lemma rtrancl_mono_subseteq_union: "A\<^sup>* \<subseteq> B\<^sup>* \<Longrightarrow> A' \<subseteq> A \<union> C \<Longrightarrow> B \<union> C \<subseteq> B' \<Longrightarrow> A'\<^sup>* \<subseteq> B'\<^sup>*"
apply (subgoal_tac "A'\<^sup>* \<subseteq> (A \<union> C)\<^sup>*") prefer 2 apply (erule rtrancl_mono)
apply (subgoal_tac "(B \<union> C)\<^sup>* \<subseteq> B'\<^sup>*") prefer 2 apply (erule rtrancl_mono)
apply (subgoal_tac "(A\<^sup>* \<union> C\<^sup>*)\<^sup>* \<subseteq> (B\<^sup>* \<union> C\<^sup>*)\<^sup>*") prefer 2 apply (rule rtrancl_mono) apply fast
apply (simp only: rtrancl_Un_rtrancl)
done

lemma interior_rel_rtrancl_mono_subseteq_union:
  " R' \<subseteq> R \<union> C 
  \<Longrightarrow> S \<union> C \<subseteq> S'
  \<Longrightarrow> (interior_rel A R)\<^sup>* \<subseteq> (interior_rel A S)\<^sup>* 
  \<Longrightarrow> (interior_rel A R')\<^sup>* \<subseteq> (interior_rel A S')\<^sup>*"
apply (erule_tac C="interior_rel A C" in rtrancl_mono_subseteq_union)
apply (rule subset_trans) apply (rule interior_rel_mono) apply (rule subset_refl) apply assumption apply simp
apply (rule_tac B="interior_rel A (S \<union> C)" in subset_trans) apply simp apply (rule interior_rel_mono) apply simp+
done


lemma interior_rel_rtrancl_mono_subseteq_union_elim:
  "(interior_rel A R)\<^sup>* \<subseteq> (interior_rel A S)\<^sup>* 
  \<Longrightarrow> R' \<subseteq> R \<union> C 
  \<Longrightarrow> S \<union> C \<subseteq> S'
  \<Longrightarrow> (interior_rel A R')\<^sup>* \<subseteq> (interior_rel A S')\<^sup>*"
apply (rule interior_rel_rtrancl_mono_subseteq_union)
apply assumption+
done


  (* trancl_mono is not named consistently with rtrancl_mono in the Isabelle library *)
lemma trancl_mono2: "r \<subseteq> s \<Longrightarrow> r\<^sup>+ \<subseteq> s\<^sup>+"
by clarsimp (fast intro: trancl_mono)


lemma trancl_mono_subseteq_union: "A\<^sup>+ \<subseteq> B\<^sup>+ \<Longrightarrow> A' = A \<union> C \<Longrightarrow> B' = B \<union> C \<Longrightarrow> A'\<^sup>+ \<subseteq> B'\<^sup>+"
apply simp
apply (subgoal_tac "(A\<^sup>+ \<union> C\<^sup>+)\<^sup>+ \<subseteq> (B\<^sup>+ \<union> C\<^sup>+)\<^sup>+")
apply (simp only: trancl_Un_trancl)
apply (rule trancl_mono2)
apply fast
done


lemma interior_rel_trancl_mono_subseteq_union:
  "(interior_rel A R)\<^sup>+ \<subseteq> (interior_rel A S)\<^sup>+
  \<Longrightarrow> R' = R \<union> C 
  \<Longrightarrow> S' = S \<union> C 
  \<Longrightarrow> (interior_rel A R')\<^sup>+ \<subseteq> (interior_rel A S')\<^sup>+"
by (erule trancl_mono_subseteq_union) simp_all


  (* The following
lemma "r\<^sup>* \<subseteq>  s\<^sup>* \<Longrightarrow> (exterior_rel A r)\<^sup>* \<subseteq> (exterior_rel A s)\<^sup>*"

does not hold: take 
r = {(y, z)}
s = {(y, a1), (a1, a2), (a2, z)}
A = {a1, a2}

then: {(y, z)}\<^sup>* \<subseteq>  {(y, a1), (a1, a2), (a2, z)}\<^sup>*, 
but not: {(y, z)}\<^sup>* \<subseteq> {(y, a1), (a2, z)}\<^sup>*,
*)

  (* The following
lemma "r\<^sup>* \<subseteq>  s\<^sup>* \<Longrightarrow> (interior_rel A r)\<^sup>* \<subseteq> (interior_rel A s)\<^sup>*"

does not hold: take 
r = {(y, z)}
s = {(y, e), (e, z)}
A = {x, y}

then: {(y, z)}\<^sup>* \<subseteq>  {(y, e), (e, z)}\<^sup>*, 
but not: {(y, z)}\<^sup>* \<subseteq> {}\<^sup>*,
*)

end


