theory Graphs imports Aux


begin

(* *** Graphs *** *)

record  'obj graph =
  nodes :: "'obj set"
  edges :: "('obj * 'obj) set"

 (*   -- "structurally well-formed graph" *)
definition
  struct_wf_gr_utpd :: "('obj, 'a) graph_scheme \<Rightarrow> bool" where
  "struct_wf_gr_utpd gr = (finite (nodes gr) \<and> (Field (edges gr)) \<subseteq> (nodes gr))"

locale graph_loc = 
  fixes gr :: "('obj, 'a) graph_scheme"

locale graph_swfgr_loc = graph_loc + 
  assumes struct_wf_gr: "struct_wf_gr_utpd gr"

context graph_swfgr_loc
begin

lemma struct_wf_gr_finite_nodes: "finite (nodes gr)"
using struct_wf_gr by (simp add: struct_wf_gr_utpd_def)

lemma struct_wf_gr_finite_edges [simp]: "finite (edges gr)"
using struct_wf_gr
by (clarsimp simp add: struct_wf_gr_utpd_def)
(blast intro: finite_rel_finite_Field)

lemma finite_edges_Image: "finite ((edges gr) `` {nn})"
proof -
have "edges gr `` {nn} \<subseteq> Field (edges gr)"
  by (fastforce simp add: Image_def Field_def)
thus ?thesis using struct_wf_gr_finite_edges
  by (blast intro: finite_subset finite_Field) 
qed


lemma finite_edges_converse_Image: "finite ((edges gr)\<inverse> `` {nn})"
proof -
have "(edges gr)\<inverse> `` {nn} \<subseteq> Field (edges gr)"
  by (fastforce simp add: Image_def converse_def Field_def)
thus ?thesis using struct_wf_gr_finite_edges
  by (blast intro: finite_subset finite_Field) 
qed

lemma struct_wf_edges_in_nodes:
  "(x, x') \<in> edges gr \<Longrightarrow> {x, x'} \<subseteq> nodes gr"
using struct_wf_gr
by (simp add: struct_wf_gr_utpd_def Field_def Domain_def Range_def) fast

lemma struct_wf_edges_clausal:
  "(x, y) \<in> edges gr \<Longrightarrow> x \<in> nodes gr \<and> y \<in> nodes gr"
by (fast dest: struct_wf_edges_in_nodes)


lemma restrict_rel_insert_wf_gr:
  "b \<notin> nodes gr \<Longrightarrow> 
  restrict_rel (edges gr) (insert b ns) = restrict_rel (edges gr) ns"
by (fastforce simp add: restrict_rel_def Field_def dest: struct_wf_edges_in_nodes)


lemma restrict_rel_edges_nodes [simp]:
  "restrict_rel (edges gr) (nodes gr) = (edges gr)"
by (fastforce simp add: restrict_rel_def Field_def dest: struct_wf_edges_in_nodes)

end -- "context graph_swfgr_loc"

end


