theory Grtransflang imports Graphtransfo_rel Transflang

(* TODO: reconsider import structure of theories. 
  Only the definition of App at the end of the theory requires Graphtransfo_rel; otherwise, on might have
theory Grtransflang imports Transflang
*)

begin

  (* Graph transformation proof system where consistency conditions 
  (well-formedness, well-typing) are handled behind the scenes *)

(* extended Hoare rules with a condition remaining constant during execution *)

definition exthoare :: "['a assn,'a assn, 'a stmt, 'a assn] => bool" ("(_) \<turnstile> {(1_)}/ (_)/ {(1_)}" 50) where
          "W \<turnstile> {P}c{Q} == \<turnstile> {\<lambda> s. W s \<and> P s} c {\<lambda> s. W s \<and> Q s}"

definition exthoare_valid :: "['a assn,'a assn, 'a stmt, 'a assn] => bool" ("(_) \<Turnstile> {(1_)}/ (_)/ {(1_)}" 50) where
          "W \<Turnstile> {P}c{Q} == \<forall> s t. (W s \<and> P s) \<longrightarrow> (\<langle>c, s\<rangle> \<Rightarrow> t) \<longrightarrow> (W t \<and> Q t)"


lemma exthoare_sound: "W \<turnstile> {P}c{Q} \<Longrightarrow> W \<Turnstile> {P}c{Q}"
apply (unfold exthoare_def)
apply (drule hoare_sound)
apply (simp add: hoare_valid_def exthoare_valid_def)
done

lemma exthoare_compl: "W \<Turnstile> {P} c {Q} \<Longrightarrow> W \<turnstile> {P} c {Q}"
apply (unfold exthoare_def)
apply (rule compl)
apply (simp add: hoare_valid_def exthoare_valid_def)
done



definition 
  wfmd_preserving :: "['a assn, 'a outcome \<Rightarrow> 'a outcome \<Rightarrow> bool] \<Rightarrow> bool" where
  "wfmd_preserving W R = (\<forall>s s'. (W s \<longrightarrow> R s s' \<longrightarrow> W s'))"


(*
lemma valid_gt_wfmd_preserving: "\<lbrakk> valid_gt (tr_transfo trr) tp \<rbrakk> \<Longrightarrow>
  wfmd_preserving (\<lambda>gr. valid_gr (outcome_val gr) tp)
     (\<lambda>gro. applicable_transfo trr (outcome_val gro) tp)
     (\<lambda>so. apply_transfo trr (outcome_val so) tp)"
apply (simp add: wfmd_preserving_def apply_transfo_def)
apply clarify
apply (rule valid_gt_valid_gr)
apply assumption+
apply (simp add: applicable_transfo_def)
apply (erule someI2_ex)
apply (simp add: applicable_transfo_gm_def)
done
*)


lemma if_raise: "(if c then a \<and> t else a \<and> e) = (a \<and> (if c then t else e))"
apply auto
done


lemma DoEH: "wfmd_preserving W R \<Longrightarrow>
  W \<turnstile> {\<lambda> s. \<forall> s'. R s s' \<longrightarrow> Q s'} (Do R) {Q}"
apply (unfold exthoare_def)
apply (rule ConseqH_P)
prefer 2 apply (rule DoH)
apply (fastforce simp add: wfmd_preserving_def)
done

lemma ThenEH: "\<lbrakk> W \<turnstile> {P} c1 {\<lambda> s. if (failed s) then (R s) else (Q s)}; 
            W \<turnstile> {Q} c2 {R} \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} (Then c1 c2) {R}"
apply (unfold exthoare_def)
apply (rule ThenH)
prefer 2 apply assumption
apply (simp add: if_raise)
done

lemma  OrelseEH: "\<lbrakk> W \<turnstile> {P} c1 {\<lambda> s. if (failed s) then (Q s) else (R s)}; 
              W \<turnstile> {Q} c2 {R} \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} (Orelse c1 c2) {R}"
apply (unfold exthoare_def)
apply (rule OrelseH)
prefer 2 apply assumption
apply (simp add: if_raise)
done

lemma ConseqEH: "\<lbrakk> 
  W \<turnstile> {P'} c {Q'}; 
  \<forall> s. W s \<longrightarrow> P s \<longrightarrow> P' s; \<forall> s. W s \<longrightarrow> Q' s \<longrightarrow> Q s \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} c {Q}"
apply (unfold exthoare_def)
apply (rule ConseqH)
prefer 2 apply assumption
apply fast+
done

lemma ConseqEH_P: "\<lbrakk> W \<turnstile> {P'} c {Q}; \<forall> s. W s \<longrightarrow> P s \<longrightarrow> P' s \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} c {Q}"
apply (rule ConseqEH)
apply assumption+
apply fast
done

lemma ConseqEH_Q: "\<lbrakk> W \<turnstile> {P} c {Q'}; \<forall> s. W s \<longrightarrow> Q' s \<longrightarrow> Q s \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} c {Q}"
apply (rule ConseqEH)
apply assumption
apply fast+
done

lemma LoopEH: "\<lbrakk> W \<turnstile> {P} c {Q}; \<forall> s. W s \<longrightarrow> Q s \<and> \<not> failed s \<longrightarrow> P s \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} (Loop c) {\<lambda> s. Q s \<and> failed s}"
apply (unfold exthoare_def)
apply (rule ConseqH_Q)
apply (rule LoopH)
apply assumption
apply fast+
done

(* Loops with invariant annotation *)
definition LoopA :: "['a assn, 'a stmt] \<Rightarrow> 'a stmt" where
  "LoopA I c = Loop c"


lemma LoopAEH: "\<lbrakk> W \<turnstile> {P} c {I};
  (\<forall>s. W s \<longrightarrow>  I s \<and> \<not> failed s \<longrightarrow> P s);
  (\<forall>s. W s \<longrightarrow> I s \<and> failed s \<longrightarrow> Q s)
 \<rbrakk> \<Longrightarrow>
     W \<turnstile> {P} (LoopA I c) {Q}"
apply (simp add: LoopA_def)
apply (rule ConseqEH_Q)
apply (rule LoopEH)
apply assumption+
done

definition
  lift :: "['a \<Rightarrow> bool, 'a \<Rightarrow> 'a \<Rightarrow> bool] \<Rightarrow> 'a outcome \<Rightarrow> 'a outcome \<Rightarrow> bool" where
  "lift b R = (\<lambda> so so'. 
               if (b (outcome_val so))
               then (\<exists> s'. (R (outcome_val so) s') \<and> so' = Scd s')
               else so' = Fld (outcome_val so))"


lemma lift_not_failed: 
  "\<not> failed s' \<Longrightarrow> lift b R s s' = (b (outcome_val s) \<and> R (outcome_val s) (outcome_val s'))"
apply (case_tac s')
apply simp_all
apply (simp add: lift_def split add: split_if)
done


lemma lift_failed:
  "failed s' \<Longrightarrow> lift b R s s' = 
  ((outcome_val s' = outcome_val s) \<and> (\<not> b (outcome_val s)))"
apply (simp add: lift_def split add: split_if)
apply (case_tac s')
apply simp_all
done



  (* TODO: the treatment of morph_prop still has to be streamlined *)
locale untyped_app_loc =
  fixes morph_prop :: "'obj graphmorph \<Rightarrow> bool"
begin

thm graph_transformation_loc.applicable_gm_def 

definition 
  applicable_transfo :: "[('nt, 'et, 'a) graphtrans_scheme, ('obj, 'a) graph_scheme] \<Rightarrow> bool" where
  "applicable_transfo gt gr = (\<exists> gm. graph_transformation_loc.applicable_gm gr (\<lambda> _. UNIV) (\<lambda> _. UNIV) morph_prop gm (appcond gt))"

thm graph_transformation_loc.apply_transfo_rel_def
definition
  "apply_transfo_rel gt gr gr' = 
  graph_transformation_loc.apply_transfo_rel gr (\<lambda> _. UNIV) (\<lambda> _. UNIV) gt (\<lambda> gm. apply_graphtrans_rel_utpd gt gm morph_prop gr) morph_prop gr'"

definition
  App :: "('nt, 'et, 'a) graphtrans_scheme \<Rightarrow> ('obj, 'a) graph_scheme stmt" where
  "App gt = Do (lift (applicable_transfo gt) (apply_transfo_rel gt))"

end -- "context untyped_app_loc"

end 
