theory Reductions  imports Graphtransfo_rel Decomposition


begin


primrec is_nodeset_set :: "'nt nodeset  \<Rightarrow> bool" where
  "is_nodeset_set Univ_set = False"
| "is_nodeset_set (Node_type_set t) = False"
| "is_nodeset_set (Singleton_set n) = True"

primrec is_edge_path_path :: "'et path \<Rightarrow> bool" where
  "is_edge_path_path Ident_pth = False"
| "is_edge_path_path Univ_pth = False" 
| "is_edge_path_path (Edge_pth n1 n2)  = True" 
| "is_edge_path_path (Edge_type_pth e) = False" (* TODO *)
| "is_edge_path_path (Compl_pth p) = False" (* TODO *)
| "is_edge_path_path (Conv_pth p) = False" (* TODO *)
| "is_edge_path_path (Seq_pth p p') = False"
| "is_edge_path_path (Alt_pth p p') = False" (* TODO: (is_edge_path_path p \<and> is_edge_path_path p') *)
| "is_edge_path_path (Clos_pth p) = False"

primrec is_edge_path_path_form :: "('nt, 'et) path_form \<Rightarrow> bool" where
  "is_edge_path_path_form (S_form n s) = True"
| "is_edge_path_path_form (P_form n n' p) = is_edge_path_path p"
| "is_edge_path_path_form (Neg_form pf) = False" (* TODO: can possibly be lifted *)
| "is_edge_path_path_form (Conj_form pf pf') = (is_edge_path_path_form pf \<and> is_edge_path_path_form pf')"
| "is_edge_path_path_form (All_form pf) = False" (* TODO: can possibly be lifted *)

(*
TODO: See routing.thy/path_form_interp and graphtransfo_rel.thy/applicable_gm
*)

primrec edge_set_path :: "'et path \<Rightarrow> (nat * nat) set" where
  "edge_set_path Ident_pth = {}"
| "edge_set_path Univ_pth =  {}"
| "edge_set_path (Edge_pth n1 n2)  = {(n1, n2)}" 
| "edge_set_path (Edge_type_pth e) = {}"
| "edge_set_path (Compl_pth et) = {}"
| "edge_set_path (Conv_pth et) = {}"
| "edge_set_path (Seq_pth p p') = {}"
| "edge_set_path (Alt_pth p p') = {}"
| "edge_set_path (Clos_pth p) = {}"


primrec edge_set_path_form :: "('nt, 'et) path_form \<Rightarrow> (nat * nat) set" where
  "edge_set_path_form (S_form n s) = {}"
| "edge_set_path_form (P_form n n' p) = edge_set_path p"
| "edge_set_path_form (Neg_form pf) = {}" (* TODO: can possibly be lifted *)
| "edge_set_path_form (Conj_form pf pf') = (edge_set_path_form pf \<union> edge_set_path_form pf')"
| "edge_set_path_form (All_form pf) = {}" (* TODO: can possibly be lifted *)


lemma graphtrans_gen_nodes_ngen_empty: "ngen gt = {} \<Longrightarrow> 
  graphtrans_gen_nodes gt gr morph_gen gen_nodes = (morph_gen = empty \<and> gen_nodes = {})"
by (auto simp add: graphtrans_gen_nodes_def)


lemma Field_fv_path_form: "struct_wf_gt_utpd gt \<Longrightarrow>
  Field (emorph gm ` edel gt) \<subseteq> (the \<circ> gm) ` fv_path_form (appcond gt)"
apply (simp add: Field_emorph)
apply (rule image_mono)
apply (simp add: struct_wf_gt_utpd_def)
done


context graph_transformation_swfgrgt_loc

begin

lemma is_edge_path_path_in_edges [rule_format]:
  "is_edge_path_path pth \<longrightarrow> path_interp I pth \<subseteq> (edges gr)"
by (induct pth) auto

  (* Precondition   is_edge_path_path_form pf   currently not required *)
lemma emorph_edge_set_path_form_edges [rule_format]:
  "path_form_interp (the \<circ> gm) pf \<longrightarrow> 
  emorph gm ` edge_set_path_form pf \<subseteq> edges gr"
apply (induct pf) apply auto apply (case_tac path) apply auto
done

  (* Precondition   is_edge_path_path_form pf   currently not required *)
lemma edge_set_path_form_incl_fv_path_form [rule_format]:
  "Field (edge_set_path_form pf) \<subseteq> fv_path_form pf"
apply (induct pf) apply auto apply (case_tac path) apply auto
done


lemma Field_egen: 
  "ngen gt = {} \<Longrightarrow> Field (egen gt) \<subseteq> fv_path_form (appcond gt)"
using struct_wf_gt
apply (simp add: struct_wf_gt_utpd_def)
apply fast
done

lemma Field_edel: "Field (edel gt) \<subseteq> fv_path_form (appcond gt)"
using struct_wf_gt
by (simp add: struct_wf_gt_utpd_def)

lemma Field_edge_set_path_form: 
  "ngen gt = {}
  \<Longrightarrow> Field (edge_set_path_form (appcond gt) - edel gt \<union> egen gt) \<subseteq> fv_path_form (appcond gt)"
using struct_wf_gt
apply (clarsimp simp add: struct_wf_gt_utpd_def)
apply (rule conjI)
apply (rule subset_trans) apply (rule Field_Diff_subseteq) apply (rule edge_set_path_form_incl_fv_path_form)
apply fast
done

  (* TODO: Also show the inverse direction of the last implication below,
  thus establishing completeness of the simplification for this fragment.
  *)
  (* Note: restriction to empty ndel and ngen sets is not a real limitation for the property under consideration, because any transformation modifying also ndel and ngen can be decomposed into a sequence of transfos that first modify the edges and then the node sets (without consequence on reachability) *)


lemma edel_egen_reduction: "ndel gt = {} \<Longrightarrow> ngen gt = {} \<Longrightarrow> 
  applicable_transfo \<and> (apply_transfo_rel gr') \<Longrightarrow>
  (edel gt)\<^sup>* \<subseteq> (egen gt)\<^sup>* \<Longrightarrow>
  (edges gr)^* \<subseteq> (edges gr')^* "
apply (erule apply_applicable_transfo_rel)

apply (simp only: apply_graphtrans_rel_def)

apply (simp add: applicable_gm_def)
apply (simp only: apply_graphtrans_rel_nodes_for_gen_def apply_graphtrans_rel_utpd_def)
apply (simp add: graphtrans_gen_nodes_ngen_empty)

apply (rule_tac A="emorph gm ` (edel gt)"  and B="emorph gm ` (egen gt)"
  and C="((edges gr) - (emorph gm `  (edel gt)))" 
  in rtrancl_mono_subseteq_union) 
apply (erule rtrancl_inclusion_emorph)

apply (fast intro: Field_edel)
apply (simp add: Field_egen)
done


end -- "context graph_transformation_swfgrgt_loc"

print_locale graph_transformation_swfgrgt_loc


locale graph_transformation_swfgrgt_inj_morph_loc = 
 graph_transformation_swfgrgt_loc _ _ _ _ _ inj_map

context graph_transformation_swfgrgt_inj_morph_loc
begin

lemma edge_set_path_form_reduction: "ndel gt = {} \<Longrightarrow> ngen gt = {} \<Longrightarrow>
  applicable_transfo \<and> (apply_transfo_rel gr') \<Longrightarrow>
  (edge_set_path_form (appcond gt) \<union> (edel gt))\<^sup>* 
  \<subseteq> ((edge_set_path_form (appcond gt) - edel gt) \<union> (egen gt))\<^sup>* \<Longrightarrow>
  (edges gr)^* \<subseteq> (edges gr')^* "
apply (erule apply_applicable_transfo_rel)

apply (simp only: apply_graphtrans_rel_def)

apply (simp add: applicable_gm_def)
apply (simp only: apply_graphtrans_rel_utpd_def 
  apply_graphtrans_rel_nodes_for_gen_def)
apply (simp add: graphtrans_gen_nodes_ngen_empty)

apply (rule_tac A="emorph gm ` (edge_set_path_form (appcond gt) \<union> edel gt)" 
  and B="emorph gm ` (edge_set_path_form (appcond gt) - edel gt \<union> egen gt)"
  and C="((edges gr) - (emorph gm `  (edel gt)))" 
  in rtrancl_mono_subseteq_union) 
apply (erule rtrancl_inclusion_emorph)

apply (simp only: image_Un)
apply fast

apply (elim conjE)
apply (subgoal_tac "emorph gm ` (edge_set_path_form (appcond gt) - edel gt)
      = (emorph gm ` edge_set_path_form (appcond gt)) - (emorph gm ` (edel gt))")
prefer 2  apply (rule inj_on_image_set_diff) apply (erule inj_on_emorph)
  apply (simp add: Field_product_subset edge_set_path_form_incl_fv_path_form) 
  apply (simp add: Field_product_subset Field_edel)

apply (subgoal_tac "emorph gm ` edge_set_path_form (appcond gt) \<subseteq> edges gr")
  prefer 2 apply (rule emorph_edge_set_path_form_edges) apply assumption+

apply (simp add: image_Un)
apply fast
done


  (* Old version having additional precondition: edel gt \<subseteq> edge_set_path_form (appcond gt) *)
lemma edge_set_path_form_reduction_old: "ndel gt = {} \<Longrightarrow> ngen gt = {} \<Longrightarrow>
  edel gt \<subseteq> edge_set_path_form (appcond gt) \<Longrightarrow>
  applicable_transfo \<and> (apply_transfo_rel gr') \<Longrightarrow>
  (edge_set_path_form (appcond gt))\<^sup>* 
  \<subseteq> ((edge_set_path_form (appcond gt) - edel gt) \<union> (egen gt))\<^sup>* \<Longrightarrow>
  (edges gr)^* \<subseteq> (edges gr')^* "
apply (erule apply_applicable_transfo_rel)

apply (simp only: apply_graphtrans_rel_def)

apply (simp add: applicable_gm_def)
apply (simp only: apply_graphtrans_rel_utpd_def 
  apply_graphtrans_rel_nodes_for_gen_def)
apply (simp add: graphtrans_gen_nodes_ngen_empty)

apply clarsimp

apply (erule_tac A="(the \<circ> gm) ` (fv_path_form (appcond gt))" in interior_exterior_decompos_rtrancl_elems)

  -- "exterior_rel"
apply (rule rtrancl_mono)
apply (simp only: exterior_rel_diff exterior_rel_union) 
apply (simp add: Field_fv_path_form Field_exterior_rel struct_wf_gt) 

  -- "interior_rel"
apply (rule_tac R="(emorph gm ` (edge_set_path_form (appcond gt)))"
  and S="emorph gm ` (edge_set_path_form (appcond gt) - edel gt \<union> egen gt)" 
  and C="((edges gr) - (emorph gm `  (edel gt)))" 
  in interior_rel_rtrancl_mono_subseteq_union)

apply (fast intro: image_mono)


apply (subgoal_tac "emorph gm ` (edge_set_path_form (appcond gt) - edel gt)
      = (emorph gm ` edge_set_path_form (appcond gt)) - (emorph gm ` (edel gt))")
prefer 2  apply (rule inj_on_image_set_diff) apply (erule inj_on_emorph)
  apply (simp add: Field_product_subset edge_set_path_form_incl_fv_path_form) 
  apply (simp add: Field_product_subset Field_edel)

apply (simp add: image_Un inj_on_image_set_diff)
apply (subgoal_tac "emorph gm ` edge_set_path_form (appcond gt) \<subseteq> edges gr")
  prefer 2 apply (rule emorph_edge_set_path_form_edges) apply assumption+
apply fast

apply (rule decompos_complete_interior)
apply (simp only: Field_emorph) apply (rule image_mono) apply (erule Field_edge_set_path_form) 
apply (erule rtrancl_inclusion_emorph)

done


end -- "context graph_transformation_swfgrgt_inj_morph_loc"


end

