theory Transflang imports Main

begin

(* OUTCOME *)



datatype 'a outcome = 
    Scd 'a    -- "succeeded"
  | Fld 'a    -- "failed"


fun  outcome_val :: "'a outcome \<Rightarrow> 'a" where
  "outcome_val (Scd v) = v"
| "outcome_val (Fld v) = v"

fun inv_outcome :: "'a outcome \<Rightarrow> 'a outcome" where
  "inv_outcome (Scd v) = Fld v"
| "inv_outcome (Fld v) = Scd v"

fun failed :: "'a outcome \<Rightarrow> bool" where
  "failed (Scd v) = False"
| "failed (Fld v) = True"

fun success :: "'a outcome \<Rightarrow> bool" where
  "success (Scd v) = True"
| "success (Fld v) = False"

lemma failed_Fld: "failed s = (\<exists> v. s = (Fld v))"
by (case_tac s) auto

lemma not_failed_Scd: "(\<not> failed s) = (\<exists> v. s = (Scd v))"
by (case_tac s) auto


datatype 'a stmt
  = Do "'a outcome \<Rightarrow> 'a outcome \<Rightarrow> bool"
  | Then "'a stmt" "'a stmt"      -- "sequential composition, fails when first fails"
  | Orelse  "'a stmt" "'a stmt"   -- "parallel composition with retry"
  | Loop  "'a stmt"               -- "loop"


(* Idioms: 
  Skip = (Do (\<lambda> x. True) (\<lambda> g. g))
  (Cond b c) = (Then (Do b (\<lambda> g. g)) c)
  (While b c) = (Loop (Then (Do b (\<lambda> g. g)) c))
  *)

(* MOVE ELSEWHERE
constdefs
  App :: "[graphtrans, graphmorph] \<Rightarrow> graph stmt"
  "App gt gm == Do (\<lambda> so. valid_application gt gm (outcome_val so)) 
                   (\<lambda> so. apply_graphtrans gt gm (outcome_val so))"
 *)

(*
Selapp gt = Do (\<lambda> gro. applicable gt gro) (\<lambda> gro. apply_graphtrans gt (sel_gm gt gro) (outcome_val gro))

  SelappS: "\<lbrakk> valid_application gt gm (outcome_val gro); 
             apply_graphtrans gt gm (outcome_val gro) = gr' \<rbrakk> \<Longrightarrow> 
  \<langle>Selapp gt, gro\<rangle> \<Rightarrow> Scd gr'"
  SelappF:  "\<lbrakk> \<not> (\<exists> gm. valid_application gt gm (outcome_val gro)) \<rbrakk> \<Longrightarrow> 
  \<langle>Selapp gt, gro\<rangle> \<Rightarrow> Fld gr"
*)


inductive exec :: "['a stmt, 'a outcome, 'a outcome] \<Rightarrow> bool" ("\<langle>_,_\<rangle>/ \<Rightarrow> _" [0,0,60] 60)
(* intros*)
where
  DoSF: "\<lbrakk> R s s' \<rbrakk> \<Longrightarrow>
  \<langle>Do R, s\<rangle> \<Rightarrow> s'"

| ThenS:  "\<lbrakk> \<langle>c1, so\<rangle> \<Rightarrow> Scd gr'; \<langle>c2, Scd gr'\<rangle> \<Rightarrow> s \<rbrakk> \<Longrightarrow> 
  \<langle>Then c1 c2, so\<rangle> \<Rightarrow> s"
| ThenF:  "\<lbrakk> \<langle>c1, so\<rangle> \<Rightarrow> Fld gr' \<rbrakk> \<Longrightarrow> 
  \<langle>Then c1 c2, so\<rangle> \<Rightarrow> Fld gr'"

| OrelseS:  "\<lbrakk> \<langle>c1, so\<rangle> \<Rightarrow> Scd gr' \<rbrakk> \<Longrightarrow> 
  \<langle>Orelse c1 c2, so\<rangle> \<Rightarrow> Scd gr'"
| OrelseF:  "\<lbrakk> \<langle>c1, so\<rangle> \<Rightarrow> Fld gr'; \<langle>c2, Fld gr'\<rangle> \<Rightarrow> s \<rbrakk> \<Longrightarrow> 
  \<langle>Orelse c1 c2, so\<rangle> \<Rightarrow> s"

| LoopS:  "\<lbrakk> \<langle>c, so\<rangle> \<Rightarrow> Scd gr'; \<langle>Loop c, Scd gr'\<rangle> \<Rightarrow> s \<rbrakk> \<Longrightarrow> 
  \<langle>Loop c, so\<rangle> \<Rightarrow> s"
| LoopT:  "\<lbrakk> \<langle>c, so\<rangle> \<Rightarrow> Fld gr' \<rbrakk> \<Longrightarrow> 
  \<langle>Loop c, so\<rangle> \<Rightarrow> Fld gr'"


inductive_cases DoSFE [elim!]: "\<langle>Do R, s\<rangle> \<Rightarrow> s'"
inductive_cases ThenE [elim!]: "(\<langle>Then c1 c2, so\<rangle> \<Rightarrow> s')"
inductive_cases OrelseE [elim!]: "(\<langle>Orelse c1 c2, so\<rangle> \<Rightarrow> s')"
inductive_cases LoopE [elim!]: "(\<langle>Loop c, so\<rangle> \<Rightarrow> s')"

lemma DoSF_simp [simp]: "(\<langle>Do R, so\<rangle> \<Rightarrow> so') = (R so so')"
by (blast intro: DoSF)


lemma Loop_unwind: "(\<langle>Loop c,s\<rangle> \<Rightarrow> t) = \<langle>Then c (Loop c),s\<rangle> \<Rightarrow> t"
apply (rule iffI)
apply (erule LoopE)
apply (rule ThenS) apply assumption+
apply (blast intro: ThenF)
apply (erule ThenE)
apply (rule LoopS) apply assumption+
apply (blast intro: LoopT) 
done


(* HOARE RULES *)

type_synonym 'a assn = "'a outcome => bool"

definition hoare_valid :: "['a assn, 'a stmt, 'a assn] => bool" ("\<Turnstile> {(1_)}/ (_)/ {(1_)}" 50)
          where "\<Turnstile> {P}c{Q} == \<forall> s t. P s \<longrightarrow> (\<langle>c, s\<rangle> \<Rightarrow> t) \<longrightarrow> Q t"

inductive hoare :: "['a assn, 'a stmt, 'a assn] \<Rightarrow> bool" ("\<turnstile> ({(1_)}/ (_)/ {(1_)})" 50)
where
  DoH: " \<turnstile> {\<lambda> s. \<forall> s'. R s s' \<longrightarrow> Q s'} (Do R) {Q}"

| ThenH: "\<lbrakk> \<turnstile> {P} c1 {\<lambda> s. if (failed s) then (R s) else (Q s)}; 
            \<turnstile> {Q} c2 {R} \<rbrakk> \<Longrightarrow>
     \<turnstile> {P} (Then c1 c2) {R}"

| OrelseH: "\<lbrakk> \<turnstile> {P} c1 {\<lambda> s. if (failed s) then (Q s) else (R s)}; 
              \<turnstile> {Q} c2 {R} \<rbrakk> \<Longrightarrow>
     \<turnstile> {P} (Orelse c1 c2) {R}"

| ConseqH: "\<lbrakk> \<forall> s. P s \<longrightarrow> P' s; \<turnstile> {P'} c {Q'}; \<forall> s. Q' s \<longrightarrow> Q s \<rbrakk> \<Longrightarrow>
     \<turnstile> {P} c {Q}"

| LoopH: "\<lbrakk> \<turnstile> {P} c {Q}; \<forall> s. Q s \<and> \<not> failed s \<longrightarrow> P s \<rbrakk> \<Longrightarrow>
     \<turnstile> {P} (Loop c) {\<lambda> s. Q s \<and> failed s}"

lemma hoare_sound_loop: "\<lbrakk>\<langle>c',s\<rangle> \<Rightarrow> t; 
  \<forall>s . P s \<longrightarrow> (\<forall> t. \<langle>c,s\<rangle> \<Rightarrow> t \<longrightarrow> Q t); 
  \<forall>s. Q s \<and> \<not> failed s \<longrightarrow> P s; P s; c' = Loop c \<rbrakk>
       \<Longrightarrow> Q t \<and> failed t"

by (induct set: exec) simp_all


lemma hoare_sound: "\<turnstile> {P}c{Q} \<Longrightarrow> \<Turnstile> {P}c{Q}"
apply (unfold hoare_valid_def)
apply (induct set: hoare)

(* Do *)
apply (fastforce split add: outcome.split)

(* Then *)
apply fastforce

(* Orelse *)
apply fastforce

(* Conseq *)
apply fast

(* Loop *)
apply (clarsimp)
apply (rule hoare_sound_loop) 
apply assumption+
apply (rule refl)
done

definition wp :: "'a stmt \<Rightarrow> 'a assn \<Rightarrow> 'a assn" where
          "wp c Q == (\<lambda> s. (\<forall> t. \<langle>c, s\<rangle> \<Rightarrow> t \<longrightarrow> Q t))"

lemma wp_Do: "(wp (Do R) Q s) = (\<forall> s'. R s s' \<longrightarrow> Q s')"
by (simp add: wp_def split del: split_if)

lemma wp_Then: "wp (Then c1 c2) Q = 
  wp c1 (\<lambda> t. if (failed t) then (Q t) else (wp c2 Q t))"
apply (unfold wp_def)
apply (simp add: fun_eq_iff )
apply safe
apply (fastforce simp add: failed_Fld intro: ThenF)
apply (fastforce simp add: not_failed_Scd intro: ThenS)
apply fastforce+ 
done

lemma wp_Orelse: "wp (Orelse c1 c2) Q = 
  wp c1 (\<lambda> t. if (failed t) then (wp c2 Q t) else (Q t))"
apply (unfold wp_def)
apply (simp add: fun_eq_iff )
apply safe 
apply (fastforce simp add: failed_Fld intro: OrelseF)
apply (fastforce simp add: not_failed_Scd  intro: OrelseS)
apply fastforce+ 
done

lemma wp_Loop_Then: "wp (Loop c) Q = 
  wp (Then c (Loop c)) Q"
by (unfold wp_def) (simp add: fun_eq_iff Loop_unwind)


lemma wp_Loop: "wp (Loop c) Q = 
  wp c (\<lambda>t. (failed t \<longrightarrow> Q t) \<and> (\<not> failed t \<longrightarrow> wp (Loop c) Q t))"
apply (rule trans)
apply (simp add: wp_Loop_Then)
apply (simp add: wp_Then)
done

lemma ConseqH_P: "\<lbrakk> \<forall> s. P s \<longrightarrow> P' s; \<turnstile> {P'} c {Q} \<rbrakk> \<Longrightarrow>
     \<turnstile> {P} c {Q}"
apply (rule ConseqH) prefer 2 apply (assumption) apply auto
done

lemma ConseqH_Q: "\<lbrakk> \<turnstile> {P} c {Q'}; \<forall> s. Q' s \<longrightarrow> Q s \<rbrakk> \<Longrightarrow>
     \<turnstile> {P} c {Q}"
apply (rule ConseqH) prefer 2 apply (assumption) apply auto
done

lemma compl_aux [rule_format]: "\<forall> Q. \<turnstile> {wp c Q} c {Q}"
apply (induct c)

apply (rule allI)
apply (simp only: wp_Do)
apply (rule DoH)

apply (rule allI)
apply (simp only: wp_Then)
apply (rule ThenH)
apply (fast, fast)

apply (rule allI)
apply (simp only: wp_Orelse)
apply (rule OrelseH) 
apply (fast, fast)

apply (rule allI)
apply (rule ConseqH_Q)
apply (rule LoopH)
apply (subst wp_Loop)
apply fast+
done

lemma compl: "\<Turnstile> {P} c {Q} \<Longrightarrow> \<turnstile> {P} c {Q}"
apply (rule ConseqH [where Q'=Q])
apply (rule_tac [2] compl_aux)
apply (fastforce simp add: hoare_valid_def wp_def)+
done

end



