theory Typing imports Graphtransfo_rel

begin


  (* Graphs *)
record ('obj, 'nt, 'et) typed_graph = "'obj graph" + 
  nodetp :: "'nt \<Rightarrow> 'obj set"
  edgetp :: "'et \<Rightarrow> ('obj * 'obj) set"

definition
  struct_wf_gr_tpd :: "('obj, 'nt, 'et, 'a) typed_graph_scheme \<Rightarrow> bool" where
  "struct_wf_gr_tpd gr = 
     (struct_wf_gr_utpd gr \<and> 
     (\<forall> nt. (nodetp gr nt) \<subseteq> (nodes gr)) \<and> (\<forall> et. (edgetp gr et) \<subseteq> (edges gr)))"

lemma struct_wf_gr_tpd_impl_utpd [simp]:
  "struct_wf_gr_tpd gr \<Longrightarrow> struct_wf_gr_utpd gr"
by (simp add: struct_wf_gr_tpd_def)

locale graph_swfgr_tpd_loc = graph_loc + 
  assumes struct_wf_gr: "struct_wf_gr_tpd gr"

sublocale graph_swfgr_tpd_loc \<subseteq> graph_swfgr_loc
using struct_wf_gr by unfold_locales simp

  (* Graph Interpretations *)
locale typed_graph_interpretation_loc
  = graph_interpretation_loc gr "(nodetp gr)" "(edgetp gr)" 
    for gr :: "('obj, 'nt, 'et, 'a) typed_graph_scheme"

  (* Graph Transformations *)

record ('nt, 'et) typed_graphtrans = "('nt, 'et) graphtrans" +
  -- "typing of generated nodes"
  ngentp :: "'nt \<Rightarrow> nat set"
  -- "mapping of edges"
  edeltp  :: "'et \<Rightarrow> (nat * nat) set"
  egentp  :: "'et \<Rightarrow> (nat * nat) set"

definition
  apply_graphtrans_rel_types_for_gen :: 
  "[('nt, 'et, 'a) typed_graphtrans_scheme, 'obj graphmorph, 
            ('obj, 'nt, 'et, 'b) typed_graph_scheme, ('obj, 'nt, 'et, 'b) typed_graph_scheme,
            'obj graphmorph, 'obj set] \<Rightarrow> bool" where
  "apply_graphtrans_rel_types_for_gen gt gm gr gr' morph_gen gen_nodes =
    (let del_nodes = ran (gm |` (ndel gt)) in
     let morph_c = gm ++ morph_gen in
     let nds = ((nodes gr) - del_nodes) \<union> gen_nodes in 
     let del_edgetp = (\<lambda> et. (emorph gm) ` ((edeltp gt et) \<union> (edel gt))) in
     let gen_edgetp = (\<lambda> et. (emorph morph_c) ` (egentp gt et)) in

  (\<forall> nt. (nodetp gr' nt = ((nodetp gr nt) \<union> ((the \<circ> morph_c) ` (ngentp gt nt)))  \<inter> nds)) \<and>
  (\<forall> et. (edgetp gr' et =  (restrict_rel (edgetp gr et) nds - (del_edgetp et)) \<union> gen_edgetp et)))"

definition
  apply_graphtrans_rel_tpd :: "[('nt, 'et, 'a) typed_graphtrans_scheme, 'obj graphmorph, 'obj graphmorph \<Rightarrow> bool,
            ('obj, 'nt, 'et, 'b) typed_graph_scheme, ('obj, 'nt, 'et, 'b) typed_graph_scheme] \<Rightarrow> bool" where
  "apply_graphtrans_rel_tpd gt gm morph_prop gr gr' =
  (\<exists> morph_gen gen_nodes.
  morph_prop (gm ++ morph_gen) \<and>
  apply_graphtrans_rel_nodes_for_gen gt gm gr gr' morph_gen gen_nodes \<and>
  apply_graphtrans_rel_types_for_gen gt gm gr gr' morph_gen gen_nodes)"


definition
  struct_wf_gt_tpd :: "('nt, 'et, 'a) typed_graphtrans_scheme \<Rightarrow> bool" where
  "struct_wf_gt_tpd gt == (struct_wf_gt_utpd gt \<and> (\<forall> et. egentp gt et \<subseteq> egen gt))"

print_locale graph_swfgr_tpd_loc


locale typed_graph_transformation_loc = 
  graph_transformation_loc gr "(nodetp gr)" "(edgetp gr)" gt apply_graphtrans_rel morph_prop
  for gr :: "('obj, 'nt, 'et, 'a) typed_graph_scheme"
  and gt :: "('nt, 'et, 'a) typed_graphtrans_scheme"
  and apply_graphtrans_rel
  and morph_prop +
  defines apply_graphtrans_rel: 
  "apply_graphtrans_rel gm gr' == apply_graphtrans_rel_tpd gt gm morph_prop gr gr'"

locale typed_graph_transformation_swfgrgt_loc = typed_graph_transformation_loc +
  assumes struct_wf_gr: "struct_wf_gr_tpd gr"
  assumes struct_wf_gt: "struct_wf_gt_tpd gt"

context typed_graph_transformation_swfgrgt_loc
begin

lemma Field_edgetp_in_nodes [simp]: "Field (edgetp gr et) \<subseteq> nodes gr"
using struct_wf_gr
apply (simp add: struct_wf_gr_tpd_def struct_wf_gr_utpd_def)
apply (elim conjE)
apply (rule subset_trans)
prefer 2 apply assumption
apply (fast intro!: mono_Field)
done

lemma restrict_rel_edgetp_nodes [simp]:
  "restrict_rel (edgetp gr et) (nodes gr) = (edgetp gr et)"
apply ( simp add: restrict_rel_def Field_def)
apply (rule)
apply (simp add: Field_product_subset)+
done


  (* Needs tyding *)

theorem struct_wf_gt_struct_wf_gr_rel_tpd: 
  assumes  "applicable_gm gm (appcond gt)"
  and "morph_sep gm (ndel gt)"
  and "apply_graphtrans_rel gm gr'"
  shows "struct_wf_gr_tpd gr'"
using assms struct_wf_gr struct_wf_gt
apply (simp (no_asm_simp) add: struct_wf_gr_tpd_def)
apply (intro conjI)
apply (rule graph_transformation_swfgrgt_loc.struct_wf_gt_struct_wf_gr_rel)
apply unfold_locales apply (clarsimp simp add: struct_wf_gr_tpd_def) apply assumption+
 apply (clarsimp simp add: struct_wf_gr_tpd_def) apply (unfold struct_wf_gt_tpd_def) apply fast apply assumption+

apply (clarsimp simp add: apply_graphtrans_rel_def apply_graphtrans_rel_tpd_def apply_graphtrans_rel_utpd_def
apply_graphtrans_rel_nodes_for_gen_def Let_def
)
apply fast


apply (clarsimp simp add: apply_graphtrans_rel_def apply_graphtrans_rel_tpd_def apply_graphtrans_rel_utpd_def
apply_graphtrans_rel_nodes_for_gen_def Let_def
)
apply (simp add: apply_graphtrans_rel_types_for_gen_def Let_def)


apply (clarsimp simp add: apply_graphtrans_rel_def apply_graphtrans_rel_tpd_def apply_graphtrans_rel_utpd_def
apply_graphtrans_rel_nodes_for_gen_def Let_def
)
apply (simp add: apply_graphtrans_rel_types_for_gen_def Let_def)
apply (erule disjE conjE)
apply (clarsimp simp add: struct_wf_gr_tpd_def image_Un)
apply (subgoal_tac "restrict_rel (edgetp gr et) (nodes gr - ran (gm |` ndel gt) \<union> gen_nodes) 
  \<subseteq> restrict_rel (edges gr) (nodes gr - ran (gm |` ndel gt) \<union> gen_nodes)")
   prefer 2 apply (rule restrict_rel_mono) apply fast apply fast apply (rule rev_subsetD) prefer 2 apply assumption+


apply (subgoal_tac "emorph (gm ++ morph_gen) ` egentp gt et \<subseteq> emorph (gm ++ morph_gen) ` egen gt")
apply (drule set_mp) apply assumption apply fast
apply (rule image_mono)
apply fast
done


end -- "context typed_graph_transformation_swfgrgt_loc"

sublocale typed_graph_transformation_loc  \<subseteq> graph_loc .

(*
sublocale typed_graph_transformation_swfgrgt_loc \<subseteq> typed_graph_swfgr_loc
using struct_wf_gr by unfold_locales 

locale typed_graph_transformation_swfgrgt_inj_morph_loc = 
 typed_graph_transformation_swfgrgt_loc _ _ _ inj_map
*)

end
